/***
*qsort.c - quicksort algorithm; qsort() library function for sorting arrays
*
*       Copyright (c) 1985-1997, Microsoft Corporation. All rights reserved.
*
*Purpose:
*       To implement the qsort() routine for sorting arrays.
*
*******************************************************************************/

#include "z1mrp.h"

static void  swap ( char *a, char *b, unsigned width)
{
    char tmp;

    if ( a != b )
        while ( width-- ) {
            tmp = *a;
            *a++ = *b;
            *b++ = tmp;
        }
}
void  zsort (ZMRP &g, void *base, int num, int width, int (*comp)(ZMRP &, const void *, const void *))
{
    char *lo, *hi;              /* ends of sub-array currently sorting */
    char *mid;                  /* points to middle of subarray */
    char *loguy, *higuy;        /* traveling pointers for partition step */
    unsigned size;              /* size of the sub-array */
    char *lostk[30], *histk[30];
    int stkptr;                 /* stack for saving sub-array to be processed */

    if (num < 2 || width == 0) return;
    stkptr = 0;                 /* initialize stack */
    lo = (char *)base;
    hi = (char *)base + width * (num-1);        /* initialize limits */
recurse:
    size = (hi - lo) / width + 1;        /* number of el's to sort */
    mid = lo + (size / 2) * width;      /* find middle element */
    swap(mid, lo, width);               /* swap it to beginning of array */
    loguy = lo;
    higuy = hi + width;
    for (;;) {
        do  {
            loguy += width;
        } while (loguy <= hi && comp(g, loguy, lo) <= 0);
        do  {
            higuy -= width;
        } while (higuy > lo && comp(g, higuy, lo) >= 0);
        if (higuy < loguy) break;
        swap(loguy, higuy, width);
    }
    swap(lo, higuy, width);     /* put partition element in place */
    if ( (higuy - lo) - 1 >= hi - loguy ) {
        if (lo + width < higuy) {
            lostk[stkptr] = lo;
            histk[stkptr] = higuy - width;
            ++stkptr;
        }                           /* save big recursion for later */
        if (loguy < hi) {
            lo = loguy;
            goto recurse;           /* do small recursion */
        }
    } else {
        if (loguy < hi) {
            lostk[stkptr] = loguy;
            histk[stkptr] = hi;
            ++stkptr;               /* save big recursion for later */
        }
        if (lo + width < higuy) {
            hi = higuy - width;
            goto recurse;           /* do small recursion */
        }
    }
    --stkptr;
    if (stkptr >= 0) {
        lo = lostk[stkptr];
        hi = histk[stkptr];
        goto recurse;           /* pop subarray from stack */
    }
    return;
}
